import ast
import pathlib
from typing import Any, Optional

import boto3
import nbformat
from nbclient import NotebookClient
from nbconvert import HTMLExporter

from firenze.exceptions import VariableAssignmentError


class Notebook:
    def __init__(
        self,
        notebook: nbformat.notebooknode.NotebookNode,
        client: Optional[NotebookClient] = None,
    ):
        if client is None:
            client = NotebookClient(notebook, timeout=600)
        self.client = client
        self.jupyter_notebook = notebook

    def execute(self, **kwargs):
        for key, value in kwargs.items():
            self.set_first_assignment_of_variable(key, value)
        with self.client.setup_kernel():
            for index, cell in enumerate(self.jupyter_notebook.cells):
                self.client.execute_cell(cell, index)

    @property
    def cells(self):
        return self.jupyter_notebook.cells

    @property
    def code_cells(self):
        return [c for c in self.cells if c["cell_type"] == "code"]

    @property
    def html(self) -> str:
        return HTMLExporter().from_notebook_node(self.jupyter_notebook)[0]

    def is_clean(self) -> bool:
        return all([c["outputs"] == [] for c in self.cells])

    def clean(self):
        for cell in self.jupyter_notebook.cells:
            cell["outputs"] = []
            try:
                del cell["execution_count"]
            except KeyError:
                pass

    @classmethod
    def from_path(cls, notebook_path, client: Optional[NotebookClient] = None):
        with open(notebook_path) as f:
            jupyter_notebook = nbformat.read(f, as_version=4)
        return cls(jupyter_notebook, client)

    def get_first_assignment_of_variable(self, variable_name: str):
        for cell in self.code_cells:
            tree = ast.parse(cell["source"])
            for node in ast.walk(tree):
                if not isinstance(node, ast.Assign):
                    continue
                for target in node.targets:
                    if isinstance(target, ast.Name) and target.id == variable_name:
                        return ast.literal_eval(node.value)

    def set_first_assignment_of_variable(self, variable_name: str, variable_value: Any):
        for cell in self.code_cells:
            try:
                tree = ast.parse(cell["source"])
            except SyntaxError:
                continue
            for node in ast.walk(tree):
                if not isinstance(node, ast.Assign):
                    continue
                for target in node.targets:
                    if isinstance(target, ast.Name) and target.id == variable_name:
                        node.value = ast.Constant(variable_value)
                        cell["source"] = ast.unparse(tree)
                        return
        raise VariableAssignmentError(
            f"Variable {variable_name} not found. Maybe in a cell with a magic command?"
        )

    @classmethod
    def from_s3(cls, s3_path, s3_client=None):
        if s3_client is None:
            s3_client = boto3.client("s3")
        bucket, key = s3_path.replace("s3://", "").split("/", 1)
        data = s3_client.get_object(Bucket=bucket, Key=key)
        jupyter_notebook = nbformat.reads(data["Body"].read().decode("utf_8"), as_version=4)
        return cls(jupyter_notebook)

    def write_html(self, file_path):
        if file_path.startswith("s3://"):
            self.write_html_to_s3(file_path)
        else:
            self.write_html_to_local(file_path)

    def write_html_to_local(self, file_path):
        pathlib.Path(file_path).parent.mkdir(parents=True, exist_ok=True)
        with open(file_path, "w") as f:
            f.write(self.html)

    def write_html_to_s3(self, s3_path):
        bucket, key = s3_path.replace("s3://", "").split("/", 1)
        s3_client = boto3.client("s3")
        s3_client.put_object(Bucket=bucket, Key=key, Body=self.html)



#!/usr/bin/env python3
import json

import click

from firenze.exceptions import VariableAssignmentError
from firenze.notebook import Notebook


@click.command()
@click.argument("notebook-path", type=click.Path(exists=True))
@click.option("--output-html-path", type=click.Path(), default="output.html")
@click.argument("parameters", nargs=-1)
def execute_notebook(notebook_path, output_html_path, parameters):
    parsed_options = parse_options(parameters)
    notebook = Notebook.from_path(notebook_path)
    notebook.clean()
    should_print = True
    try:
        notebook.execute(**parsed_options)
    except Exception as e:
        should_print = not isinstance(e, VariableAssignmentError)
        raise
    finally:
        if should_print:
            notebook.write_html(output_html_path)


def parse_options(parameters):
    parsed_options = {}
    for option in parameters:
        name, value = option.split("=", 1)
        try:
            value = json.loads(value)
        except json.JSONDecodeError:
            pass
        parsed_options[name] = value
    return parsed_options


if __name__ == "__main__":
    execute_notebook()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['firenze']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.26.156,<2.0.0',
 'click>=8.1.3,<9.0.0',
 'ipykernel>=6.22.0,<7.0.0',
 'nbclient>=0.7.4,<0.8.0',
 'nbconvert>=7.3.1,<8.0.0',
 'nbformat>=5.8.0,<6.0.0']

entry_points = \
{'console_scripts': ['firenze = firenze.cli:execute_notebook']}

setup_kwargs = {
    'name': 'firenze',
    'version': '0.1.1',
    'description': 'A lean executor for jupyter notebooks.',
    'long_description': '# Firenze\n\nFirenze is a lean jupyter notebook executor, that generates the notebook output in a single HTML\nfile. You can also parameterize the notebooks without any modification to the notebook itself.\nIt supports local files and `s3` paths, both for the notebook and for the output.\n\n## As a Library\nYou can use `firenze` as a library in your own project. Install it through `pip`\n\n```bash\npip install firenze\n```\n\nSuppose you have a very simple notebook that runs a "Hello, World!"\n\n![A notebook in jupyter](https://github.com/pabloalcain/firenze/blob/main/docs/img/hello_world_in_jupyter.png?raw=true)\n\nYou can execute it right away with `firenze` through\n```bash\nfirenze docs/notebooks/hello_world.ipynb\n```\nand the output html will be, as expected:\n\n![Hello, World! output](https://github.com/pabloalcain/firenze/blob/main/docs/img/hello_world_output.png?raw=true)\n\nYou can also send parameters and `firenze` will automatically modify the variable:\n\n```bash\nfirenze docs/notebooks/hello_world.ipynb name=Firenze\n```\n\n![Hello, Firenze! output](https://github.com/pabloalcain/firenze/blob/main/docs/img/hello_world_with_parameters.png?raw=true)\n\n## As a Docker Image\nThis is still in the making, but one idea is to call `firenze` as a docker image with a notebook\nand a `requirements.txt`, so the notebook execution can be easily deployed to remote servers.\n',
    'author': 'Pablo Alcain',
    'author_email': 'pabloalcain@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
